/*
 *  Aquamarine the KDE window decorator
 *
 *  Copyright (c) 2006 Dennis Kasprzyk <onestone@beryl-project.org>
 *  Copyright (c) 2006 Volker Krause <vkrause@kde.org>
 *
 *  Uses code of:
 *      Emerald window decorator (www.beryl-project.org)
 *      gtk-window-decorator (www.freedesktop.org/wiki/Software/Compiz)
 *      KWin window manager (www.kde.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include "window.h"
#include "decorator.h"
#include "options.h"
#include "utils.h"
#include "settings.h"

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/shape.h>
#include <X11/extensions/Xrender.h>

#include <math.h>

#include <fixx11h.h>

#include <kdebug.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kdecoration.h>
#include <kwin.h>
#include <klocale.h>
#include <kprocess.h>
#include <kstandarddirs.h>

#include <netwm.h>

#include <qapplication.h>
#include <qlayout.h>
#include <qevent.h>
#include <qpainter.h>
#include <qobjectlist.h>
#include <qwidget.h>
#include <qstring.h>
#include <qtimer.h>
#include <qcursor.h>
#include <qpopupmenu.h>

#define BERYL_ADVANCED_DECORATION_VERSION 10121980

#define GRAVITY_WEST  (1 << 0)
#define GRAVITY_EAST  (1 << 1)
#define GRAVITY_NORTH (1 << 2)
#define GRAVITY_SOUTH (1 << 3)

#define XX_MASK (1 << 12)
#define XY_MASK (1 << 13)
#define YX_MASK (1 << 14)
#define YY_MASK (1 << 15)

#define CLAMP_HORZ (1 << 0)
#define CLAMP_VERT (1 << 1)

#define WM_MOVERESIZE_SIZE_TOPLEFT      0
#define WM_MOVERESIZE_SIZE_TOP          1
#define WM_MOVERESIZE_SIZE_TOPRIGHT     2
#define WM_MOVERESIZE_SIZE_RIGHT        3
#define WM_MOVERESIZE_SIZE_BOTTOMRIGHT  4
#define WM_MOVERESIZE_SIZE_BOTTOM       5
#define WM_MOVERESIZE_SIZE_BOTTOMLEFT   6
#define WM_MOVERESIZE_SIZE_LEFT         7
#define WM_MOVERESIZE_MOVE              8
#define WM_MOVERESIZE_SIZE_KEYBOARD     9
#define WM_MOVERESIZE_MOVE_KEYBOARD    10

extern Atom net_window_decor;
extern Atom net_wm_moveresize;

extern Atom wm_protocols;
extern Atom wm_delete_window;
extern Atom wm_take_focus;
extern Atom net_wm_context_help;
extern Atom net_wm_ping;
extern Atom net_wm_take_activity;

extern struct _cursor cursors[3][3];

Aquamarine::Window::Window (WId wid, WId frame):
QWidget(0,0,Qt::WX11BypassWM),
m_wid (wid),
m_active (false),
m_deco (0),
m_fakeMaximized (false),
m_shapeSet (false),
m_popup (0),
m_advanced_popup (0),
m_viewport_menu (0),
m_appearance_menu (0),
m_brightness_menu (0), m_saturation_menu (0), m_opacity_menu (0),
mProcessKiller (0)
{
     KWin::WindowInfo wInf = KWin::windowInfo (m_wid, NET::WMState, 0);
	m_state = wInf.state ();

	m_name = KWin::readNameProperty (m_wid, XA_WM_NAME);

	m_icons = QIconSet (KWin::icon (m_wid, 16, 16, TRUE),
					KWin::icon (m_wid, 32, 32, TRUE));

	m_opacity = readWindowPropertyDefault (m_wid, Atoms::opacity, 0xffff);
	m_brightness =
		readWindowPropertyDefault (m_wid, Atoms::brightness, 0xffff);
	m_saturation =
		readWindowPropertyDefault (m_wid, Atoms::saturation, 0xffff);

	setBackgroundColor(QColor(0,0));

	double op = Settings::shadowOpacity() / 100.0;
	op *= 0xffff;

	m_sradius = Settings::shadowRadius();
	QColor color = Settings::shadowColor();
	m_scolor[0] = short((color.red() / 256.0) * 0xffff);
	m_scolor[1] = short((color.green() / 256.0) * 0xffff);
	m_scolor[2] = short((color.blue() / 256.0) * 0xffff);
	m_scolor[3] = short(op);

	int offX = Settings::shadowOffsetX();
	int offY = Settings::shadowOffsetY();
	offX = KMAX(-m_sradius,KMIN(m_sradius,offX));
	offY = KMAX(-m_sradius,KMIN(m_sradius,offY));

	if (Settings::drawShadows())
		m_soffset.setCoords(-m_sradius + offX,-m_sradius + offY,
							m_sradius + offX,m_sradius + offY);
	else
		m_soffset.setCoords(0,0,0,0);

	m_currentPixmap = None;
	m_updatePixmap = true;
	m_mapped = true;

	setWState (Qt::WState_Visible);
	updateFrame (frame);

	int x, y;
	unsigned int width, height, border, depth;
	installEventFilter(this);

	::Window root;
	XGetGeometry (qt_xdisplay (), m_wid, &root, &x, &y, &width, &height,
					&border, &depth);
	m_geometry = QRect (x, y, width, height);

	getWindowProtocols ();
	createDecoration ();
}

Aquamarine::Window::~Window ()
{
    delete m_deco;
    delete m_popup;

    Atom my_atom = Atoms::net_window_decor;

    Aquamarine::trapXError ();
    XWindowAttributes
        attr;
    if (XGetWindowAttributes (qt_xdisplay (), m_wid, &attr)
        && !Aquamarine::popXError ())
    {
        XDeleteProperty (qt_xdisplay (), m_wid, my_atom);
    }

	if (m_currentPixmap != None)
		XFreePixmap (qt_xdisplay(), m_currentPixmap);


    Aquamarine::popXError ();
}



bool
Aquamarine::Window::isActive () const
{
    return m_active;
}

bool
Aquamarine::Window::isCloseable () const
{
    return true;
}

bool
Aquamarine::Window::isMaximizable () const
{
    return true;
}

KDecoration::MaximizeMode Aquamarine::Window::maximizeMode () const
{
	if (m_fakeMaximized)
        return MaximizeVertical | MaximizeHorizontal;

    MaximizeMode cur = ((m_state & NET::MaxVert)
        ? MaximizeVertical : MaximizeRestore) |
            ((m_state & NET::MaxHoriz) ? MaximizeHorizontal : MaximizeRestore);
    return cur;
}

bool
Aquamarine::Window::isMinimizable () const
{
    return true;
}

bool
Aquamarine::Window::providesContextHelp () const
{
    return m_supportContextHelp;
}

int
Aquamarine::Window::desktop () const
{
    return (m_state & NET::Sticky) ? NET::OnAllDesktops : 1;
}

bool
Aquamarine::Window::isModal () const
{
    return m_state & NET::Modal;
}

bool
Aquamarine::Window::isShadeable () const
{
    return true;
}

bool
Aquamarine::Window::isShade () const
{
    return (m_state & NET::Shaded);
}

bool
Aquamarine::Window::isSetShade () const
{
    return isShade ();
}

bool
Aquamarine::Window::keepAbove () const
{
    return (m_state & NET::KeepAbove);
}

bool
Aquamarine::Window::keepBelow () const
{
    return (m_state & NET::KeepBelow);
}

bool
Aquamarine::Window::isMovable () const
{
    return true;
}

NET::WindowType Aquamarine::Window::windowType (unsigned long) const
{
    return NET::Normal;
}

bool
Aquamarine::Window::isResizable () const
{
    return true;
}

QIconSet
Aquamarine::Window::icon () const
{
    return m_icons;
}

QString
Aquamarine::Window::caption () const
{
    return m_name;
}

void
Aquamarine::Window::showWindowMenu (QPoint pos)
{
    kdDebug () << k_funcinfo << pos << endl;

    // create menu if we don't have it already
    if (!m_popup)
    {
        m_popup = new QPopupMenu (this);
        m_popup->setCheckable (true);
        m_popup->setFont (KGlobalSettings::menuFont ());

        m_advanced_popup = new QPopupMenu (m_popup);
        m_advanced_popup->setCheckable (true);
        m_advanced_popup->setFont (KGlobalSettings::menuFont ());
        connect (m_advanced_popup, SIGNAL (activated (int)),
                 SLOT (slotPopupActivated (int)));

        m_advanced_popup->insertItem (SmallIconSet ("up"),
                                      i18n ("Keep &Above Others"),
                                      Options::KeepAboveOp);
        m_advanced_popup->insertItem (SmallIconSet ("down"),
                                      i18n ("Keep &Below Others"),
                                      Options::KeepBelowOp);
        m_advanced_popup->insertItem (SmallIconSet ("window_fullscreen"),
                                      i18n ("&Fullscreen"),
                                      Options::FullScreenOp);

        m_popup->insertItem (i18n ("Ad&vanced"), m_advanced_popup);

        m_viewport_menu = new QPopupMenu (m_popup);
        m_viewport_menu->setCheckable (true);
        m_viewport_menu->setFont (KGlobalSettings::menuFont ());
        connect (m_viewport_menu, SIGNAL (activated (int)),
                 SLOT (slotChangeViewport (int)));

        m_viewport_menu->insertItem (i18n ("On all Viewports"), 0);
        m_viewport_menu->insertSeparator ();
        for (int i = 1;
             i <=
             Decorator::viewports ().height () *
             Decorator::viewports ().width (); ++i)
            m_viewport_menu->insertItem (i18n ("Move to Viewport %1").arg (i),
                                         i);

        m_popup->insertItem (i18n ("To Viewport"), m_viewport_menu);

        m_popup->insertItem (SmallIconSet ("move"), i18n ("&Move"),
                             Options::MoveOp);
        m_popup->insertItem (i18n ("Re&size"), Options::ResizeOp);

        m_popup->insertItem (i18n ("Mi&nimize"), Options::MinimizeOp);
        m_popup->insertItem (i18n ("Ma&ximize"), Options::MaximizeOp);
        m_popup->insertItem (i18n ("Sh&ade"), Options::ShadeOp);


        m_appearance_menu = new QPopupMenu (m_popup);
        m_appearance_menu->setFont (KGlobalSettings::menuFont ());

        m_opacity_menu = new QPopupMenu (m_appearance_menu);
        m_opacity_menu->setFont (KGlobalSettings::menuFont ());
        m_opacity_menu->insertItem ("25%", ChangeOpacity25);
        m_opacity_menu->insertItem ("50%", ChangeOpacity50);
        m_opacity_menu->insertItem ("75%", ChangeOpacity75);
        m_opacity_menu->insertItem ("100%", ChangeOpacity100);

        m_brightness_menu = new QPopupMenu (m_appearance_menu);
        m_brightness_menu->setFont (KGlobalSettings::menuFont ());
        m_brightness_menu->insertItem ("25%", ChangeBrightness25);
        m_brightness_menu->insertItem ("50%", ChangeBrightness50);
        m_brightness_menu->insertItem ("75%", ChangeBrightness75);
        m_brightness_menu->insertItem ("100%", ChangeBrightness100);

        m_saturation_menu = new QPopupMenu (m_appearance_menu);
        m_saturation_menu->setFont (KGlobalSettings::menuFont ());
        m_saturation_menu->insertItem ("25%", ChangeSaturation25);
        m_saturation_menu->insertItem ("50%", ChangeSaturation50);
        m_saturation_menu->insertItem ("75%", ChangeSaturation75);
        m_saturation_menu->insertItem ("100%", ChangeSaturation100);

        m_appearance_menu->insertItem (i18n ("Opacity"), m_opacity_menu);
        m_appearance_menu->insertItem (i18n ("Brightness"),
                                       m_brightness_menu);
        m_appearance_menu->insertItem (i18n ("Saturation"),
                                       m_saturation_menu);
        m_appearance_menu->insertSeparator ();
        m_appearance_menu->insertItem (i18n ("Reset settings"), ResetValues);

        m_popup->insertItem (i18n ("Appearance"), m_appearance_menu);

        connect (m_appearance_menu, SIGNAL (activated (int)),
                 SLOT (slotAppearanceChange (int)));
        connect (m_opacity_menu, SIGNAL (activated (int)),
                 SLOT (slotAppearanceChange (int)));
        connect (m_brightness_menu, SIGNAL (activated (int)),
                 SLOT (slotAppearanceChange (int)));
        connect (m_saturation_menu, SIGNAL (activated (int)),
                 SLOT (slotAppearanceChange (int)));


        m_popup->insertSeparator ();
        m_popup->insertItem (SmallIconSet ("fileclose"), i18n ("&Close"),
                             Options::CloseOp);
        m_popup->insertSeparator ();

        connect (m_popup, SIGNAL (aboutToShow ()),
                 SLOT (slotPopupAboutToShow ()));
        connect (m_popup, SIGNAL (activated (int)),
                 SLOT (slotPopupActivated (int)));
    }
    m_popup->exec (pos);
}

void
Aquamarine::Window::showWindowMenu (const QRect & pos)
{
    showWindowMenu (pos.bottomLeft ());
}

void
Aquamarine::Window::processMousePressEvent (QMouseEvent * qme)
{
    Options::MouseCommand com = Options::MouseNothing;
    bool active = isActive();
    if ( !m_supportTakeFocus )
        active = TRUE;

    switch (qme->button ())
    {
        case Qt::LeftButton:
            com = active ? Decorator::options()->commandActiveTitlebar1() : Decorator::options()->commandInactiveTitlebar1();
            break;
        case Qt::MidButton:
            com = active ? Decorator::options()->commandActiveTitlebar2() : Decorator::options()->commandInactiveTitlebar2();
            break;
        case Qt::RightButton:
            com = active ? Decorator::options()->commandActiveTitlebar3() : Decorator::options()->commandInactiveTitlebar3();
            break;
        default:
            break;
    }
    if( qme->button () == Qt::LeftButton
        && com != Options::MouseOperationsMenu // actions where it's not possible to get the matching
        && com != Options::MouseMinimize )  // mouse release event
        {
        moveWindow (qme);
    }


    performMouseCommand( com, qme);
}

void
Aquamarine::Window::performWindowOperation (WindowOperation wo)
{
    kdDebug () << k_funcinfo << wo << endl;
    switch (wo)
    {
    case KDecoration::MaximizeOp:
        maximize (maximizeMode () == KDecoration::MaximizeFull
                  ? KDecoration::MaximizeRestore : KDecoration::MaximizeFull);
        break;
    case KDecoration::HMaximizeOp:
        maximize (maximizeMode () ^ KDecoration::MaximizeHorizontal);
        break;
    case KDecoration::VMaximizeOp:
        maximize (maximizeMode () ^ KDecoration::MaximizeVertical);
        break;
    case KDecoration::MinimizeOp:
        minimize ();
        break;
    case KDecoration::ShadeOp:
        setShade (!isShade ());
        break;
    case KDecoration::CloseOp:
        closeWindow ();
        break;
    case KDecoration::KeepAboveOp:
        setKeepAbove (!keepAbove ());
        break;
    case KDecoration::KeepBelowOp:
        setKeepBelow (!keepBelow ());
        break;
    case KDecoration::FullScreenOp:
        if (m_state & NET::FullScreen)
        {
            KWin::clearState (m_wid, NET::FullScreen);
        }
        else
            KWin::setState (m_wid, NET::FullScreen);
        break;
    case KDecoration::MoveOp:
    {
		XEvent ev;
        ev.xclient.type = ClientMessage;
        ev.xclient.display = qt_xdisplay ();

        ev.xclient.serial = 0;
        ev.xclient.send_event = TRUE;

        ev.xclient.window = m_wid;
        ev.xclient.message_type = Atoms::net_wm_moveresize;
        ev.xclient.format = 32;

        ev.xclient.data.l[0] = m_geometry.x () + (m_geometry.width () / 2);
        ev.xclient.data.l[1] = m_geometry.y () + (m_geometry.height () / 2);
        ev.xclient.data.l[2] = WM_MOVERESIZE_MOVE_KEYBOARD;
        ev.xclient.data.l[3] = 1;
        ev.xclient.data.l[4] = 1;

        XUngrabPointer (qt_xdisplay (), CurrentTime);
        XUngrabKeyboard (qt_xdisplay (), CurrentTime);

        XSendEvent (qt_xdisplay (), qt_xrootwin (), FALSE,
                    SubstructureRedirectMask | SubstructureNotifyMask, &ev);

        XSync (qt_xdisplay (), FALSE);
    }
        break;
    case KDecoration::ResizeOp:
    {
        XEvent
            ev;
        ev.xclient.type = ClientMessage;
        ev.xclient.display = qt_xdisplay ();

        ev.xclient.serial = 0;
        ev.xclient.send_event = TRUE;

        ev.xclient.window = m_wid;
        ev.xclient.message_type = Atoms::net_wm_moveresize;
        ev.xclient.format = 32;

        ev.xclient.data.l[0] = m_geometry.x () + (m_geometry.width () / 2);
        ev.xclient.data.l[1] = m_geometry.y () + (m_geometry.height () / 2);
        ev.xclient.data.l[2] = WM_MOVERESIZE_SIZE_KEYBOARD;
        ev.xclient.data.l[3] = 1;
        ev.xclient.data.l[4] = 1;

        XUngrabPointer (qt_xdisplay (), CurrentTime);
        XUngrabKeyboard (qt_xdisplay (), CurrentTime);

        XSendEvent (qt_xdisplay (), qt_xrootwin (), FALSE,
                    SubstructureRedirectMask | SubstructureNotifyMask, &ev);

        XSync (qt_xdisplay (), FALSE);
    }
        break;
    default:
        break;
    }
}

void
Aquamarine::Window::setMask (const QRegion & reg, int)
{
    kdDebug () << k_funcinfo << endl;

	int top, bottom, left, right;
    m_deco->borders (left, right, top, bottom);

	if (m_shapeSet)
		m_shapeChange += m_shape.eor(reg);
	else
		m_shapeChange = reg;

    m_shape = QRegion (reg);
    m_shapeSet = true;

	if (m_deco)
		m_deco->widget()->setMask(reg);

	QRegion moved = QRegion (reg);
	moved.translate(-m_soffset.left(),-m_soffset.top());

    XShapeCombineRegion (qt_xdisplay (), handle(), ShapeInput, 0, 0,
                         moved.handle (), ShapeSet);

	//renderShadow();
}

bool
Aquamarine::Window::isPreview () const
{
    return false;
}

QRect
Aquamarine::Window::geometry () const
{
    if (m_deco)
    {
        int top, bottom, left, right;
        m_deco->borders (left, right, top, bottom);
        return QRect (0, 0, m_geometry.width () + left + right,
                      m_geometry.height () + top + bottom);
    }
    return m_geometry;
}

QRect
Aquamarine::Window::iconGeometry () const
{
    return QRect ();
}

QRegion
Aquamarine::Window::unobscuredRegion (const QRegion & r) const
{
    return r;
}

QWidget *
Aquamarine::Window::workspaceWidget () const
{
    return 0;
}

WId
Aquamarine::Window::windowId () const
{
    return m_wid;
}

void
Aquamarine::Window::closeWindow ()
{
    kdDebug () << k_funcinfo << endl;
    if (!isCloseable ())
        return;
    if (m_supportDeleteWindow)
    {
        Aquamarine::Decorator::sendClientMessage (m_wid, Atoms::wm_protocols,
                                                  Atoms::wm_delete_window);
    }
    else
    {
        XKillClient (qt_xdisplay (), m_wid);
    }
}

void
Aquamarine::Window::maximize (MaximizeMode mode)
{
    kdDebug () << k_funcinfo << endl;

    KWin::setState (m_wid,
                    ((mode & MaximizeVertical) ? NET::
                     MaxVert : 0) | ((mode & MaximizeHorizontal) ? NET::
                                     MaxHoriz : 0));
    KWin::clearState (m_wid,
                      ((mode & MaximizeVertical) ? 0 : NET::
                       MaxVert) | ((mode & MaximizeHorizontal) ? 0 : NET::
                                   MaxHoriz));
}

void
Aquamarine::Window::minimize ()
{
    kdDebug () << k_funcinfo << endl;
    KWin::iconifyWindow (m_wid, false);
}

void
Aquamarine::Window::showContextHelp ()
{
    kdDebug () << k_funcinfo << endl;
    if (m_supportContextHelp)
    {
        Aquamarine::Decorator::sendClientMessage (m_wid, Atoms::wm_protocols,
                                                  Atoms::net_wm_context_help);
    }
}

void
Aquamarine::Window::titlebarDblClickOperation ()
{
    kdDebug () << k_funcinfo << endl;
    performWindowOperation (Aquamarine::Decorator::options ()->
                            operationTitlebarDblClick ());
}

void
Aquamarine::Window::setDesktop (int desktop)
{
    kdDebug () << k_funcinfo << endl;
    if (desktop == NET::OnAllDesktops)
    {
        KWin::setState (m_wid, NET::Sticky);
    }
    else
    {
        KWin::clearState (m_wid, NET::Sticky);
    }
}

void
Aquamarine::Window::setKeepBelow (bool set)
{
    if (set)
    {
        KWin::clearState (m_wid, NET::KeepAbove);
        KWin::setState (m_wid, NET::KeepBelow);
    }
    else
        KWin::clearState (m_wid, NET::KeepBelow);
}

void
Aquamarine::Window::setKeepAbove (bool set)
{
    if (set)
    {
        KWin::clearState (m_wid, NET::KeepBelow);
        KWin::setState (m_wid, NET::KeepAbove);
    }
    else
        KWin::clearState (m_wid, NET::KeepAbove);
}

void
Aquamarine::Window::setShade (bool set)
{
    kdDebug () << k_funcinfo << endl;
    if (set)
        KWin::setState (m_wid, NET::Shaded);
    else
        KWin::clearState (m_wid, NET::Shaded);
    m_deco->shadeChange ();
}

void
Aquamarine::Window::titlebarMouseWheelOperation (int delta)
{
    kdDebug () << k_funcinfo << endl;
    performMouseCommand( Decorator::options()->operationTitlebarMouseWheel( delta ), 0);
}


int
Aquamarine::Window::currentDesktop () const
{
    return 1;
}

QWidget *
Aquamarine::Window::initialParentWidget () const
{
    return const_cast < Window * >(this);
}

Qt::WFlags Aquamarine::Window::initialWFlags () const
{
    return 0;
}

void
Aquamarine::Window::helperShowHide (bool)
{
}

void
Aquamarine::Window::grabXServer (bool)
{
}

void
Aquamarine::Window::createDecoration ()
{
    if (m_deco)
        return;
    KDecoration *
        deco = Decorator::pluginManager ()->createDecoration (this);
    deco->init ();
	deco->widget()->move(-m_soffset.left(),-m_soffset.top());
    m_deco = deco;
	m_deco->widget()->installEventFilter(this);
    resizeDecoration ();

	updateFrame(m_frame);
}

void
Aquamarine::Window::resizeDecoration ()
{
    int top, bottom, left, right;
    m_deco->borders (left, right, top, bottom);

	  kdDebug () << k_funcinfo << endl;

	m_updatePixmap = true;

	move (m_geometry.x() - left + m_soffset.left(),
		  m_geometry.y() - top + m_soffset.top());

	if (m_geometry.width () + left + right + m_soffset.width() != width() ||
		m_geometry.height () + top + bottom + m_soffset.height() != height())
	{
		m_shapeSet = false;
		resize (m_geometry.width () + left + right + m_soffset.width(),
				m_geometry.height () + top + bottom + m_soffset.height());

		m_deco->resize (QSize(m_geometry.width () + left + right,
								m_geometry.height () + top + bottom));
	}
	m_deco->widget ()->setShown (true);
}

void
Aquamarine::Window::updateWindowProperties ()
{

	Pixmap pix = XCompositeNameWindowPixmap(qt_xdisplay(),handle());

    int top, bottom, left, right;
    int mtop, mbottom, mleft, mright;

    m_fakeMaximized = true;
    m_deco->borders (mleft, mright, mtop, mbottom);
    m_fakeMaximized = false;
    m_deco->borders (left, right, top, bottom);

    int w_w = m_geometry.width ();
    int w_h = m_geometry.height ();

	int otop = top - m_soffset.top();
	int obottom = bottom + m_soffset.bottom();
	int oleft = left - m_soffset.left();
	int oright = right + m_soffset.right();

    long mdata[256];

    long *data = mdata;

    *data++ = BERYL_ADVANCED_DECORATION_VERSION;

    memcpy (data++, &pix, sizeof (pix));

    *data++ = left;
    *data++ = right;
    *data++ = top;
    *data++ = bottom;

    *data++ = mleft;
    *data++ = mright;
    *data++ = mtop;
    *data++ = mbottom;

    *data++ = 0;
    *data++ = 0;

	*data++ = 0;
	*data++ = (unsigned short)((Settings::activeOpacity() / 100.0) * 0xffff);
	*data++ = (unsigned short)((Settings::inactiveOpacity() / 100.0) * 0xffff);

    // TOP LEFT
    *data++ =
        ((GRAVITY_NORTH | GRAVITY_WEST) << 0) |
        ((GRAVITY_NORTH | GRAVITY_WEST) << 4) |
        (0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

    *data++ = -oleft;
    *data++ = -otop;
    *data++ = 0;
    *data++ = 0;
    *data++ = oleft;
    *data++ = otop;
    *data++ = 0;
    *data++ = 0;

	// TOP RIGHT
	*data++ =
        ((GRAVITY_NORTH | GRAVITY_EAST) << 0) |
        ((GRAVITY_NORTH | GRAVITY_EAST) << 4) |
        (0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

    *data++ = 0;
    *data++ = -otop;
    *data++ = oright;
    *data++ = 0;
    *data++ = oright;
    *data++ = otop;
    *data++ = oleft + w_w;
    *data++ = 0;

	// BOTTOM LEFT
    *data++ =
        ((GRAVITY_SOUTH | GRAVITY_WEST) << 0) |
        ((GRAVITY_SOUTH | GRAVITY_WEST) << 4) |
        (0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

    *data++ = -oleft;
    *data++ = 0;
    *data++ = 0;
    *data++ = obottom;
    *data++ = oleft;
    *data++ = obottom;
    *data++ = 0;
    *data++ = w_h + otop;

	// BOTTOM RIGHT
	*data++ =
        ((GRAVITY_SOUTH | GRAVITY_EAST) << 0) |
        ((GRAVITY_SOUTH | GRAVITY_EAST) << 4) |
        (0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

    *data++ = 0;
    *data++ = 0;
    *data++ = oright;
    *data++ = obottom;
    *data++ = oright;
    *data++ = obottom;
    *data++ = oleft + w_w;
    *data++ = w_h + otop;

	// TOP CENTER
	*data++ =
		((GRAVITY_NORTH | GRAVITY_WEST) << 0) |
		((GRAVITY_NORTH | GRAVITY_EAST) << 4) |
		(0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

	*data++ = 0;
	*data++ = -otop;
	*data++ = 0;
	*data++ = 0;
	*data++ = w_w;
	*data++ = otop;
	*data++ = oleft;
	*data++ = 0;

	*data++ =
		((GRAVITY_NORTH | GRAVITY_WEST) << 0) |
		((GRAVITY_NORTH | GRAVITY_EAST) << 4) |
		(0 << 8) | (CLAMP_HORZ << 10) | (YY_MASK);

	*data++ = w_w;
	*data++ = -otop;
	*data++ = 0;
	*data++ = 0;
	*data++ = SHRT_MAX;
	*data++ = otop;
	*data++ = oleft + w_w;
	*data++ = 0;

	// BOTTOM CENTER
	*data++ =
		((GRAVITY_SOUTH | GRAVITY_WEST) << 0) |
		((GRAVITY_SOUTH | GRAVITY_EAST) << 4) |
		(0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

	*data++ = 0;
	*data++ = 0;
	*data++ = 0;
	*data++ = obottom;
	*data++ = w_w;
	*data++ = obottom;
	*data++ = oleft;
	*data++ = w_h + otop;

	*data++ =
		((GRAVITY_SOUTH | GRAVITY_WEST) << 0) |
		((GRAVITY_SOUTH | GRAVITY_EAST) << 4) |
		(0 << 8) | (CLAMP_HORZ << 10) | (YY_MASK);

	*data++ = w_w;
	*data++ = 0;
	*data++ = 0;
	*data++ = obottom;
	*data++ = SHRT_MAX;
	*data++ = obottom;
	*data++ = oleft + w_w;
	*data++ = w_h + otop;

	//LEFT
	*data++ =
		((GRAVITY_NORTH | GRAVITY_WEST) << 0) |
		((GRAVITY_SOUTH | GRAVITY_WEST) << 4) |
		(0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

	*data++ = -oleft;
	*data++ = 0;
	*data++ = 0;
	*data++ = 0;
	*data++ = oleft;
	*data++ = w_h;
	*data++ = 0;
	*data++ = otop;

	*data++ =
		((GRAVITY_NORTH | GRAVITY_WEST) << 0) |
		((GRAVITY_SOUTH | GRAVITY_WEST) << 4) |
		(0 << 8) | (CLAMP_VERT << 10) | (XX_MASK);

	*data++ = -oleft;
	*data++ = w_h;
	*data++ = 0;
	*data++ = 0;
	*data++ = oleft;
	*data++ = SHRT_MAX;
	*data++ = 0;
	*data++ = otop + w_h;

	//RIGHT
    *data++ =
        ((GRAVITY_NORTH | GRAVITY_EAST) << 0) |
        ((GRAVITY_SOUTH | GRAVITY_EAST) << 4) |
        (0 << 8) | (0 << 10) | (XX_MASK) | (YY_MASK);

    *data++ = 0;
    *data++ = 0;
    *data++ = oright;
    *data++ = 0;
    *data++ = oright;
    *data++ = w_h;
    *data++ = w_w + oleft;
    *data++ = otop;

	*data++ =
        ((GRAVITY_NORTH | GRAVITY_EAST) << 0) |
        ((GRAVITY_SOUTH | GRAVITY_EAST) << 4) |
        (0 << 8) | (CLAMP_VERT << 10) | (XX_MASK);

    *data++ = 0;
    *data++ = w_h;
    *data++ = oright;
    *data++ = 0;
    *data++ = oright;
    *data++ = SHRT_MAX;
    *data++ = w_w + oleft;
    *data++ = otop + w_h;

    Aquamarine::trapXError ();
    XChangeProperty (qt_xdisplay (), m_wid,
                     Atoms::net_window_decor,
                     XA_INTEGER,
                     32, PropModeReplace, (unsigned char *)mdata,
                     15 + (12 * 9));
    Aquamarine::popXError ();
    XSync (qt_xdisplay (), FALSE);
	if (m_currentPixmap != None)
		XFreePixmap(qt_xdisplay(),m_currentPixmap);
	m_currentPixmap = pix;
}

void
Aquamarine::Window::setActive (bool active)
{
    m_active = active;
    m_deco->activeChange ();
    resizeDecoration ();
}

void
Aquamarine::Window::updateFrame (WId frame)
{
    m_frame = frame;
	if (m_deco && frame != handle())
	{
		XEvent ev;

		memset (&ev, 0, sizeof (ev));
		ev.xclient.type = ClientMessage;
		ev.xclient.window = m_wid;
		ev.xclient.message_type = Atoms::change_frame_window;
		ev.xclient.format = 32;
		ev.xclient.data.l[0] = handle();
		ev.xclient.data.l[1] = TRUE;
		ev.xclient.data.l[2] = 0;
		ev.xclient.data.l[3] = 0;
		ev.xclient.data.l[4] = 0;

		XSendEvent (qt_xdisplay (), qt_xrootwin(), False,
				SubstructureRedirectMask | SubstructureNotifyMask, &ev);
	}
	if (m_deco && frame == handle())
	{
		//updateWindowProperties ();
		XSelectInput (qt_xdisplay (), m_frame,
				defaultMask | ButtonPressMask | ButtonReleaseMask |
				PointerMotionMask | EnterWindowMask | LeaveWindowMask);
	}
}

void
Aquamarine::Window::updateWindowGeometry ()
{
    int x, y;
    unsigned int width, height, border, depth;
    ::Window root;
    XGetGeometry (qt_xdisplay (), m_wid, &root, &x, &y, &width, &height,
                  &border, &depth);
    m_geometry = QRect (x, y, width, height);
    resizeDecoration ();
}

void
Aquamarine::Window::updateConfig ()
{
	if (!m_deco) return;

	double op = Settings::shadowOpacity() / 100.0;
	op *= 0xffff;

	m_sradius = Settings::shadowRadius();
	QColor color = Settings::shadowColor();
	m_scolor[0] = short((color.red() / 256.0) * 0xffff);
	m_scolor[1] = short((color.green() / 256.0) * 0xffff);
	m_scolor[2] = short((color.blue() / 256.0) * 0xffff);
	m_scolor[3] = short(op);

	int offX = Settings::shadowOffsetX();
	int offY = Settings::shadowOffsetY();
	offX = KMAX(-m_sradius,KMIN(m_sradius,offX));
	offY = KMAX(-m_sradius,KMIN(m_sradius,offY));

	m_deco->widget()->move(m_soffset.left(),m_soffset.top());

	if (Settings::drawShadows())
		m_soffset.setCoords(-m_sradius + offX,-m_sradius + offY,
							m_sradius + offX,m_sradius + offY);
	else
		m_soffset.setCoords(0,0,0,0);

	m_deco->widget()->move(-m_soffset.left(),-m_soffset.top());

	resizeDecoration();
	updateWindowProperties();

	if (m_shapeSet)
	{
		m_shapeChange = m_shape;
		renderShadow();
	}
}

void
Aquamarine::Window::reloadDecoration ()
{
    delete m_deco;
    m_deco = 0;

    createDecoration ();
}

void
Aquamarine::Window::updateCursor (QPoint pos)
{
    switch (m_deco->mousePosition (pos))
    {
    case PositionCenter:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[1][1].cursor);
        break;
    case PositionLeft:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[1][0].cursor);
        break;
    case PositionRight:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[1][2].cursor);
        break;
    case PositionTop:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[0][1].cursor);
        break;
    case PositionBottom:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[2][1].cursor);
        break;
    case PositionTopLeft:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[0][0].cursor);
        break;
    case PositionTopRight:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[0][2].cursor);
        break;
    case PositionBottomLeft:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[2][0].cursor);
        break;
    case PositionBottomRight:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[2][2].cursor);
        break;
    default:
        XDefineCursor (qt_xdisplay (), m_frame, cursors[1][1].cursor);
        break;
    }
}

void
Aquamarine::Window::getWindowProtocols ()
{
    Atom *p;
    int i, n;

    m_supportDeleteWindow = false;
    m_supportTakeFocus = false;
    m_supportTakeActivity = false;
    m_supportContextHelp = false;
    m_supportPing = false;

    if (XGetWMProtocols (qt_xdisplay (), m_wid, &p, &n))
    {
        for (i = 0; i < n; i++)
            if (p[i] == Atoms::wm_delete_window)
                m_supportDeleteWindow = true;
            else if (p[i] == Atoms::wm_take_focus)
                m_supportTakeFocus = true;
            else if (p[i] == Atoms::net_wm_take_activity)
                m_supportTakeActivity = true;
            else if (p[i] == Atoms::net_wm_context_help)
                m_supportContextHelp = true;
            else if (p[i] == Atoms::net_wm_ping)
                m_supportPing = true;
        if (n > 0)
            XFree (p);
    }
}

void
Aquamarine::Window::slotPopupActivated (int id)
{
    WindowOperation
        op = static_cast < WindowOperation > (id);
    performWindowOperation (op);
}

void
Aquamarine::Window::slotPopupAboutToShow ()
{
    m_popup->setItemEnabled (Options::ResizeOp, isResizable ());
    m_popup->setItemEnabled (Options::MoveOp, isMovable ());

    m_popup->setItemEnabled (Options::MaximizeOp, isMaximizable ());
    m_popup->setItemChecked (Options::MaximizeOp,
                             maximizeMode () == MaximizeFull);

    m_popup->setItemChecked (Options::ShadeOp, isShade ());
    m_popup->setItemEnabled (Options::ShadeOp, isShadeable ());

    m_advanced_popup->setItemChecked (Options::KeepAboveOp, keepAbove ());
    m_advanced_popup->setItemChecked (Options::KeepBelowOp, keepBelow ());
    m_advanced_popup->setItemChecked (Options::FullScreenOp, m_state & NET::FullScreen);

    m_popup->setItemEnabled (Options::MinimizeOp, isMinimizable ());
    m_popup->setItemEnabled (Options::CloseOp, isCloseable ());


    m_opacity_menu->setItemEnabled (ChangeOpacity25, m_opacity != 0x3fff);
    m_opacity_menu->setItemEnabled (ChangeOpacity50, m_opacity != 0x7fff);
    m_opacity_menu->setItemEnabled (ChangeOpacity75, m_opacity != 0xbfff);
    m_opacity_menu->setItemEnabled (ChangeOpacity100, m_opacity != 0xffff);

    m_brightness_menu->setItemEnabled (ChangeBrightness25,
                                       m_brightness != 0x3fff);
    m_brightness_menu->setItemEnabled (ChangeBrightness50,
                                       m_brightness != 0x7fff);
    m_brightness_menu->setItemEnabled (ChangeBrightness75,
                                       m_brightness != 0xbfff);
    m_brightness_menu->setItemEnabled (ChangeBrightness100,
                                       m_brightness != 0xffff);

    m_saturation_menu->setItemEnabled (ChangeSaturation25,
                                       m_saturation != 0x3fff);
    m_saturation_menu->setItemEnabled (ChangeSaturation50,
                                       m_saturation != 0x7fff);
    m_saturation_menu->setItemEnabled (ChangeSaturation75,
                                       m_saturation != 0xbfff);
    m_saturation_menu->setItemEnabled (ChangeSaturation100,
                                       m_saturation != 0xffff);

    m_appearance_menu->setItemEnabled (ResetValues, m_opacity != 0xffff
                                       || m_brightness != 0xffff
                                       || m_saturation != 0xffff);

    m_viewport_menu->setItemChecked (0, desktop () == NET::OnAllDesktops);
    for (int i = 1;
         i <=
         Decorator::viewports ().width () * Decorator::viewports ().height ();
         ++i)
    {
        if (Decorator::
            onViewport (QPoint (m_geometry.x (), m_geometry.y ())) == i)
        {
            m_viewport_menu->setItemEnabled (i, false);
        }
        else
        {
            m_viewport_menu->setItemEnabled (i, true);
        }
    }
}

void
Aquamarine::Window::updateState ()
{
    KWin::WindowInfo wInf = KWin::windowInfo (m_wid, NET::WMState, 0);
    unsigned long
        newState = wInf.state ();
    unsigned long
        stateChange = m_state ^ newState;
    m_state = newState;

    if (stateChange & NET::Max)
    {
        m_deco->maximizeChange ();
        resizeDecoration ();
    }
    if (stateChange & NET::KeepAbove && !(m_state & NET::KeepAbove))
        m_deco->emitKeepAboveChanged (m_state & NET::KeepAbove);
    if (stateChange & NET::KeepBelow && !(m_state & NET::KeepBelow))
        m_deco->emitKeepBelowChanged (m_state & NET::KeepBelow);
    if (stateChange & NET::KeepAbove && m_state & NET::KeepAbove)
        m_deco->emitKeepAboveChanged (m_state & NET::KeepAbove);
    if (stateChange & NET::KeepBelow && m_state & NET::KeepBelow)
        m_deco->emitKeepBelowChanged (m_state & NET::KeepBelow);
    if (stateChange & NET::Shaded)
        m_deco->shadeChange ();
    if (stateChange & NET::Sticky)
        m_deco->desktopChange ();
}

void
Aquamarine::Window::updateName ()
{
    m_name = KWin::readNameProperty (m_wid, XA_WM_NAME);
    m_deco->captionChange ();
}

void
Aquamarine::Window::updateIcons ()
{
    kdDebug () << k_funcinfo << endl;
    m_icons =
        QIconSet (KWin::icon (m_wid, 16, 16, TRUE),
                  KWin::icon (m_wid, 32, 32, TRUE));
    m_deco->iconChange ();
}

void
Aquamarine::Window::slotChangeViewport (int id)
{
    if (id == 0)
        setDesktop ((desktop () ==
                     NET::OnAllDesktops) ? 1 : NET::OnAllDesktops);
    else
        Decorator::moveToViewport (m_wid, id);
}

void
Aquamarine::Window::slotAppearanceChange (int id)
{
    switch (id)
    {
    case ResetValues:
        setWindowProperty (m_wid, Atoms::opacity, 0xffff);
        setWindowProperty (m_wid, Atoms::brightness, 0xffff);
        setWindowProperty (m_wid, Atoms::saturation, 0xffff);
        break;
    case ChangeOpacity25:
        setWindowProperty (m_wid, Atoms::opacity, 0x3fff);
        break;
    case ChangeOpacity50:
        setWindowProperty (m_wid, Atoms::opacity, 0x7fff);
        break;
    case ChangeOpacity75:
        setWindowProperty (m_wid, Atoms::opacity, 0xbfff);
        break;
    case ChangeOpacity100:
        setWindowProperty (m_wid, Atoms::opacity, 0xffff);
        break;
    case ChangeBrightness25:
        setWindowProperty (m_wid, Atoms::brightness, 0x3fff);
        break;
    case ChangeBrightness50:
        setWindowProperty (m_wid, Atoms::brightness, 0x7fff);
        break;
    case ChangeBrightness75:
        setWindowProperty (m_wid, Atoms::brightness, 0xbfff);
        break;
    case ChangeBrightness100:
        setWindowProperty (m_wid, Atoms::brightness, 0xffff);
        break;
    case ChangeSaturation25:
        setWindowProperty (m_wid, Atoms::saturation, 0x3fff);
        break;
    case ChangeSaturation50:
        setWindowProperty (m_wid, Atoms::saturation, 0x7fff);
        break;
    case ChangeSaturation75:
        setWindowProperty (m_wid, Atoms::saturation, 0xbfff);
        break;
    case ChangeSaturation100:
        setWindowProperty (m_wid, Atoms::saturation, 0xffff);
        break;
    default:
        break;
    }
}

void
Aquamarine::Window::moveWindow (QMouseEvent * qme)
{
    if (!qme)
        return;
    int
        direction = WM_MOVERESIZE_MOVE;
    switch (m_deco->mousePosition (qme->pos ()))
    {

    case PositionCenter:
        direction = WM_MOVERESIZE_MOVE;
        break;
    case PositionLeft:
        direction = WM_MOVERESIZE_SIZE_LEFT;
        break;
    case PositionRight:
        direction = WM_MOVERESIZE_SIZE_RIGHT;
        break;
    case PositionTop:
        direction = WM_MOVERESIZE_SIZE_TOP;
        break;
    case PositionBottom:
        direction = WM_MOVERESIZE_SIZE_BOTTOM;
        break;
    case PositionTopLeft:
        direction = WM_MOVERESIZE_SIZE_TOPLEFT;
        break;
    case PositionTopRight:
        direction = WM_MOVERESIZE_SIZE_TOPRIGHT;
        break;
    case PositionBottomLeft:
        direction = WM_MOVERESIZE_SIZE_BOTTOMLEFT;
        break;
    case PositionBottomRight:
        direction = WM_MOVERESIZE_SIZE_BOTTOMRIGHT;
        break;
    default:
        direction = WM_MOVERESIZE_MOVE;
        break;
    }

    XEvent
        ev;
    ev.xclient.type = ClientMessage;
    ev.xclient.display = qt_xdisplay ();

    ev.xclient.serial = 0;
    ev.xclient.send_event = TRUE;

    ev.xclient.window = m_wid;
    ev.xclient.message_type = Atoms::net_wm_moveresize;
    ev.xclient.format = 32;

    ev.xclient.data.l[0] = qme->globalX ();
    ev.xclient.data.l[1] = qme->globalY ();
    ev.xclient.data.l[2] = direction;
    ev.xclient.data.l[3] = qme->button ();
    ev.xclient.data.l[4] = 1;

    XUngrabPointer (qt_xdisplay (), CurrentTime);
    XUngrabKeyboard (qt_xdisplay (), CurrentTime);

    XSendEvent (qt_xdisplay (), RootWindow (qt_xdisplay (), qt_xscreen ()),
                FALSE, SubstructureRedirectMask | SubstructureNotifyMask,
                &ev);

    XSync (qt_xdisplay (), FALSE);
}

void Aquamarine::Window::slotDesktopGeometryChanged()
{
    // force menu regeneration
    delete m_popup;
    m_popup = 0;
}


void Aquamarine::Window::performMouseCommand(Options::MouseCommand command, QMouseEvent *qme)
{
    switch (command)
    {
        case Options::MouseRaise:
        case Options::MouseActivateAndRaise:
            KWin::raiseWindow (m_wid);
            break;
        case Options::MouseLower:
            KWin::lowerWindow (m_wid);
            break;
        case Options::MouseShade :
            setShade (!isShade ());
            break;
        case Options::MouseSetShade:
            setShade (true);
            break;
        case Options::MouseUnsetShade:
            setShade (false);
            break;
        case Options::MouseOperationsMenu:
            {
                QPoint mp(0,0);
                if (qme)
                    mp = QPoint(qme->globalX (), qme->globalY ());
                showWindowMenu (mp);
                break;
            }
        case Options::MouseMaximize:
            maximize( KDecoration::MaximizeFull );
            break;
        case Options::MouseRestore:
            maximize( KDecoration::MaximizeRestore );
            break;
        case Options::MouseMinimize:
            minimize();
            break;
        case Options::MouseAbove:
            {
                if( keepBelow())
                    setKeepBelow( false );
                else
                    setKeepAbove( true );
                break;
            }
        case Options::MouseBelow:
            {
                if( keepAbove())
                    setKeepAbove( false );
                else
                    setKeepBelow( true );
                break;
            }
        case Options::MousePreviousDesktop:
            {
                int vp = Decorator::onViewport (m_geometry.topLeft ());
                Decorator::moveToViewport (m_wid, vp-1);
                break;
            }
        case Options::MouseNextDesktop:
            {
                int vp = Decorator::onViewport (m_geometry.topLeft ());
                Decorator::moveToViewport (m_wid, vp+1);
                break;
            }
        case Options::MouseOpacityMore:
            if (m_opacity < 0xFFFF)
            {
                if (m_opacity < 0xF333)
                    setWindowProperty (m_wid, Atoms::opacity, m_opacity + 0xCCC);
                else
                    setWindowProperty (m_wid, Atoms::opacity, 0xFFFF);
            }
            break;
        case Options::MouseOpacityLess:
            if (m_opacity > 0)
                setWindowProperty (m_wid, Atoms::opacity, (m_opacity > 0xCCC)? m_opacity - 0xCCC : 0);
            break;
        case Options::MouseActivateRaiseAndMove:
        case Options::MouseActivateRaiseAndUnrestrictedMove:
        case Options::MouseMove:
        case Options::MouseUnrestrictedMove:
        case Options::MouseResize:
        case Options::MouseUnrestrictedResize:
            moveWindow (qme);
            break;
        case Options::MouseNothing:
        default:
            break;
    }
}

void Aquamarine::Window::setMapping (bool map)
{
	m_mapped = map;
	if (!m_mapped)
		m_shapeSet = false;
}

void Aquamarine::Window::updateWindow ()
{
	if (!m_mapped) return;

	if (m_updatePixmap)
	{
		updateWindowProperties();
		m_updatePixmap = false;
	}
	renderShadow();
}

#if INT_MAX != LONG_MAX
/* XRenderSetPictureFilter used to be broken on LP64. This
 * works with either the broken or fixed version.
 */
static void
XRenderSetPictureFilter_wrapper (Display *dpy,
				 Picture picture,
				 char    *filter,
				 XFixed  *params,
				 int     nparams)
{
    Aquamarine::trapXError ();
    XRenderSetPictureFilter (dpy, picture, filter, params, nparams);
    XSync (dpy, False);
    if (Aquamarine::popXError ())
    {
		long *long_params = new long[nparams];
		int  i;

		for (i = 0; i < nparams; i++)
			long_params[i] = params[i];

		XRenderSetPictureFilter (dpy, picture, filter,
					 (XFixed *) long_params, nparams);
		delete long_params;
    }
}


#define XRenderSetPictureFilter XRenderSetPictureFilter_wrapper
#endif

#define SIGMA(r) ((r) / 2.0)
#define ALPHA(r) (r)

void Aquamarine::Window::renderShadow ()
{
	if (m_shapeChange.isEmpty() || !Settings::drawShadows()) return;
	kdDebug () << k_funcinfo << endl;

    int top, bottom, left, right;
    m_deco->borders (left, right, top, bottom);

	renderShadow(m_shapeChange.intersect(QRegion(0,0,m_deco->width(),top)));
	renderShadow(m_shapeChange.intersect(QRegion(0,m_deco->height() - bottom,m_deco->width(),bottom)));
	renderShadow(m_shapeChange.intersect(QRegion(0,top,left,m_geometry.height())));
	renderShadow(m_shapeChange.intersect(QRegion(m_deco->width() - right,top,right,m_geometry.height())));

	m_shapeChange = QRegion();
}

void Aquamarine::Window::renderShadow (QRegion reg)
{

	if (reg.isEmpty())
		return;

	int s2 = m_sradius * 2;

	QRect rect = reg.boundingRect();

	rect.setLeft(rect.left() - s2);
	rect.setTop(rect.top() - s2);
	rect.setRight(rect.right() + s2);
	rect.setBottom(rect.bottom() + s2);

	QRegion bound(rect);
	bound = m_shape.intersect(bound);

	bound.translate(-rect.left(),-rect.top());

	QPixmap qRegPix(rect.width(), rect.height(), 32);
	QPixmap qTmpPix(rect.width() + s2, rect.height(), 32);
	qRegPix.fill( QColor(0,0));

	QPainter p(&qRegPix);
	QBrush b(QColor(0, ((m_scolor[3] >> 8) << 24) + 0xffffff));
	p.setClipRegion(bound);
	p.fillRect(0, 0, rect.width(), rect.height(), b);

	XFixed *params;
	int n_params = 0;
	int size;

	params = createGaussianKernel (&size);

	n_params = size + 2;
	size = size / 2;

	params[0] = (n_params - 2) << 16;
	params[1] = 1 << 16;

	XRenderPictFormat *format;
	format = XRenderFindStandardFormat (qt_xdisplay(), PictStandardARGB32);

	char *filter = NULL;
	XFilters* filters = XRenderQueryFilters (qt_xdisplay(), qRegPix.handle());
	if (filters)
	{
		int i;

		for (i = 0; i < filters->nfilter; i++)
		{
			if (strcmp (filters->filter[i], FilterConvolution) == 0)
			{
				filter = (char *) FilterConvolution;
				break;
			}
		}

		XFree (filters);
	}
	if (!filter)
		return;

	static XRenderColor white = { 0xffff, 0xffff, 0xffff, 0xffff };

	Picture fill = XRenderCreateSolidFill (qt_xdisplay(), &white);
	Picture src = XRenderCreatePicture(qt_xdisplay(), qRegPix.handle(), format, 0, NULL);
	Picture tmp = XRenderCreatePicture(qt_xdisplay(), qTmpPix.handle(), format, 0, NULL);
	Picture dst = XRenderCreatePicture(qt_xdisplay(), handle(), format, 0, NULL);

	setPictureTransform(src,size,0);
	XRenderSetPictureFilter (qt_xdisplay(), src, filter, params, n_params);
	XRenderComposite (qt_xdisplay(), PictOpSrc, fill, src, tmp, 0, 0, 0, 0, 0, 0, rect.width() + s2, rect.height());

	XRenderFreePicture (qt_xdisplay(), fill);

	params[0] = 1 << 16;
    params[1] = (n_params - 2) << 16;

	XRenderFreePicture (qt_xdisplay(), src);
	XRenderColor color = { m_scolor[0], m_scolor[1], m_scolor[2], 0xffff };

	src = XRenderCreateSolidFill (qt_xdisplay(), &color);

	setPictureTransform(tmp,0,size);
	XRenderSetPictureFilter (qt_xdisplay(), tmp, filter, params, n_params);

	XRenderComposite (qt_xdisplay(), PictOpSrc, src, tmp, dst, s2, s2, s2, s2, rect.left() + s2, rect.top() + s2, rect.width() - s2, rect.height() - s2);

	XRenderFreePicture (qt_xdisplay(), src);
	XRenderFreePicture (qt_xdisplay(), tmp);
	XRenderFreePicture (qt_xdisplay(), dst);

	delete params;
}

XFixed *Aquamarine::Window::createGaussianKernel (int *r_size)
{
	double sigma = m_sradius / 2.0;
	double alpha = m_sradius;
	double radius = m_sradius;

    XFixed *params;
    double *amp, scale, x_scale, fx, sum;
    int    size, half_size, x, i, n;

    scale = 1.0f / (2.0f * M_PI * sigma * sigma);
    half_size = int(alpha + 0.5f);

    if (half_size == 0)
	half_size = 1;

    size = half_size * 2 + 1;
    x_scale = 2.0f * radius / size;

    if (size < 3)
	return NULL;

    n = size;

    amp = new double[n];
    if (!amp)
	return NULL;

    n += 2;

    params = new XFixed[n];
    if (!params)
	return NULL;

    i   = 0;
    sum = 0.0f;

    for (x = 0; x < size; x++)
    {
	fx = x_scale * (x - half_size);

	amp[i] = scale * exp ((-1.0f * (fx * fx)) / (2.0f * sigma * sigma));

	sum += amp[i];

	i++;
    }

    /* normalize */
    if (sum != 0.0)
	sum = 1.0 / sum;

    params[0] = params[1] = 0;

    for (i = 2; i < n; i++)
	params[i] = XDoubleToFixed (amp[i - 2] * sum);

    delete amp;

    *r_size = size;

    return params;
}

void Aquamarine::Window::setPictureTransform (Picture p, int dx, int dy)
{
	XTransform transform = {
		{
			{ 1 << 16, 0,       -dx << 16 },
			{ 0,       1 << 16, -dy << 16 },
			{ 0,       0,         1 << 16 },
		}
	};

	XRenderSetPictureTransform (qt_xdisplay(), p, &transform);
}

bool
Aquamarine::Window::eventFilter (QObject * watched, QEvent * event)
{
	bool rv = QWidget::eventFilter (watched, event);
    if (event->type () == QEvent::Paint)
    {
		QTimer::singleShot (0, this, SLOT(updateWindow()));
    }

    return rv;
}

void Aquamarine::Window::handleProcessKillerExited (void)
{
    if (mProcessKiller)
    {
		delete mProcessKiller;
		mProcessKiller = NULL;
    }
}

void Aquamarine::Window::showKillProcessDialog (Time timestamp)
{
    KWin::WindowInfo kWinInfo =
	KWin::windowInfo (m_wid, 0, NET::WM2WindowClass | NET::WM2ClientMachine);
    NETWinInfo       wInfo = NETWinInfo (qt_xdisplay(), m_wid,
					 qt_xrootwin (), NET::WMPid);
    QCString	     clientMachine, resourceClass;
    pid_t	     pid;
    char	     buf[257];

    if (mProcessKiller)
		return;

    clientMachine = kWinInfo.clientMachine ();
    resourceClass = kWinInfo.windowClassClass ();
    pid		  = wInfo.pid ();

    if (gethostname (buf, sizeof (buf) - 1) == 0)
    {
		if (strcmp (buf, clientMachine) == 0)
			clientMachine = "localhost";
    }

    mProcessKiller = new KProcess (this);

    *mProcessKiller << KStandardDirs::findExe ("kwin_killer_helper") <<
	"--pid" << QCString ().setNum (pid) <<
	"--hostname" << clientMachine <<
	"--windowname" << m_name.utf8 () <<
	"--applicationname" << resourceClass <<
	"--wid" << QCString ().setNum (m_wid) <<
	"--timestamp" << QCString ().setNum (timestamp);

    connect (mProcessKiller, SIGNAL (processExited (KProcess *)),
			 SLOT (handleProcessKillerExited ()));

    if (!mProcessKiller->start (KProcess::NotifyOnExit))
    {
		delete mProcessKiller;
		mProcessKiller = NULL;
    }
}

void Aquamarine::Window::hideKillProcessDialog (void)
{
    handleProcessKillerExited ();
}
